@@ -8,6 +8,7 @@ import re
 import shutil
 import subprocess
 import sys
+import tempfile
 import threading
 from collections import OrderedDict
 from concurrent.futures import ThreadPoolExecutor
@@ -15,11 +16,27 @@ from datetime import datetime
 from io import StringIO
 from multiprocessing import cpu_count
 from pathlib import Path
-from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Union
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    TypeVar,
+    Union,
+)
 
+import pandas as pd
 from tqdm.auto import tqdm
 
-from cmdstanpy import _CMDSTAN_REFRESH, _CMDSTAN_SAMPLING, _CMDSTAN_WARMUP
+from cmdstanpy import (
+    _CMDSTAN_REFRESH,
+    _CMDSTAN_SAMPLING,
+    _CMDSTAN_WARMUP,
+    _TMPDIR,
+)
 from cmdstanpy.cmdstan_args import (
     CmdStanArgs,
     GenerateQuantitiesArgs,
@@ -51,6 +68,7 @@ from cmdstanpy.utils import (
 from . import progress as progbar
 
 OptionalPath = Union[str, os.PathLike, None]
+Fit = TypeVar('Fit', CmdStanMCMC, CmdStanMLE, CmdStanVB)
 
 
 class CmdStanModel:
@@ -169,9 +187,14 @@ class CmdStanModel:
             if not cmdstan_version_before(
                 2, 27
             ):  # unknown end of version range
-                model_info = self.src_info()
-                if 'parameters' in model_info:
-                    self._fixed_param |= len(model_info['parameters']) == 0
+                try:
+                    model_info = self.src_info()
+                    if 'parameters' in model_info:
+                        self._fixed_param |= len(model_info['parameters']) == 0
+                except ValueError as e:
+                    if compile:
+                        raise
+                    get_logger().debug(e)
 
         if exe_file is not None:
             self._exe_file = os.path.realpath(os.path.expanduser(exe_file))
@@ -269,32 +292,22 @@ class CmdStanModel:
         If stanc is older than 2.27 or if the stan
         file cannot be found, returns an empty dictionary.
         """
-        result: Dict[str, Any] = {}
-        if self.stan_file is None:
-            return result
-        try:
-            cmd = (
-                [os.path.join(cmdstan_path(), 'bin', 'stanc' + EXTENSION)]
-                # handle include-paths, allow-undefined etc
-                + self._compiler_options.compose_stanc()
-                + [
-                    '--info',
-                    str(self.stan_file),
-                ]
+        if self.stan_file is None or cmdstan_version_before(2, 27):
+            return {}
+        cmd = (
+            [os.path.join(cmdstan_path(), 'bin', 'stanc' + EXTENSION)]
+            # handle include-paths, allow-undefined etc
+            + self._compiler_options.compose_stanc()
+            + ['--info', str(self.stan_file)]
+        )
+        proc = subprocess.run(cmd, capture_output=True, text=True, check=False)
+        if proc.returncode:
+            raise ValueError(
+                f"Failed to get source info for Stan model "
+                f"'{self._stan_file}'. Console:\n{proc.stderr}"
             )
-            proc = subprocess.run(
-                cmd, capture_output=True, text=True, check=True
-            )
-            result = json.loads(proc.stdout)
-            return result
-        except (
-            ValueError,
-            RuntimeError,
-            OSError,
-            subprocess.CalledProcessError,
-        ) as e:
-            get_logger().debug(e)
-            return result
+        result: Dict[str, Any] = json.loads(proc.stdout)
+        return result
 
     def format(
         self,
@@ -1200,7 +1213,7 @@ class CmdStanModel:
     def generate_quantities(
         self,
         data: Union[Mapping[str, Any], str, os.PathLike, None] = None,
-        mcmc_sample: Union[CmdStanMCMC, List[str], None] = None,
+        previous_fit: Union[Fit, List[str], None] = None,
         seed: Optional[int] = None,
         gq_output_dir: OptionalPath = None,
         sig_figs: Optional[int] = None,
@@ -1208,14 +1221,18 @@ class CmdStanModel:
         refresh: Optional[int] = None,
         time_fmt: str = "%Y%m%d%H%M%S",
         timeout: Optional[float] = None,
-    ) -> CmdStanGQ:
+        *,
+        mcmc_sample: Union[CmdStanMCMC, List[str], None] = None,
+    ) -> CmdStanGQ[Fit]:
         """
         Run CmdStan's generate_quantities method which runs the generated
         quantities block of a model given an existing sample.
 
-        This function takes a :class:`CmdStanMCMC` object and the dataset used
-        to generate that sample and calls to the CmdStan ``generate_quantities``
-        method to generate additional quantities of interest.
+        This function takes one of the Stan fit objects
+        :class:`CmdStanMCMC`, :class:`CmdStanMLE`, or :class:`CmdStanVB`
+        and the data required for the model and calls to the CmdStan
+        ``generate_quantities`` method to generate additional quantities of
+        interest.
 
         The :class:`CmdStanGQ` object records the command, the return code,
         and the paths to the generate method output CSV and console files.
@@ -1234,9 +1251,10 @@ class CmdStanModel:
             either as a dictionary with entries matching the data variables,
             or as the path of a data file in JSON or Rdump format.
 
-        :param mcmc_sample: Can be either a :class:`CmdStanMCMC` object returned
-            by the :meth:`sample` method or a list of stan-csv files generated
-            by fitting the model to the data using any Stan interface.
+        :param previous_fit: Can be either a :class:`CmdStanMCMC`,
+            :class:`CmdStanMLE`, or :class:`CmdStanVB` or a list of
+            stan-csv files generated by fitting the model to the data
+            using any Stan interface.
 
         :param seed: The seed for random number generator. Must be an integer
             between 0 and 2^32 - 1. If unspecified,
@@ -1270,39 +1288,64 @@ class CmdStanModel:
 
         :return: CmdStanGQ object
         """
-        if isinstance(mcmc_sample, CmdStanMCMC):
-            mcmc_fit = mcmc_sample
-            sample_csv_files = mcmc_sample.runset.csv_files
-        elif isinstance(mcmc_sample, list):
-            if len(mcmc_sample) < 1:
+        if mcmc_sample is not None:
+            if previous_fit:
+                raise ValueError(
+                    "Cannot supply both 'previous_fit' and "
+                    "deprecated argument 'mcmc_sample'"
+                )
+            get_logger().warning(
+                "Argument name `mcmc_sample` is deprecated, please "
+                "rename to `previous_fit`."
+            )
+
+            previous_fit = mcmc_sample  # type: ignore
+
+        if isinstance(previous_fit, (CmdStanMCMC, CmdStanMLE, CmdStanVB)):
+            fit_object = previous_fit
+            fit_csv_files = previous_fit.runset.csv_files
+        elif isinstance(previous_fit, list):
+            if len(previous_fit) < 1:
                 raise ValueError(
                     'Expecting list of Stan CSV files, found empty list'
                 )
             try:
-                sample_csv_files = mcmc_sample
-                sample_fit = from_csv(sample_csv_files)
-                mcmc_fit = sample_fit  # type: ignore
+                fit_csv_files = previous_fit
+                fit_object = from_csv(fit_csv_files)  # type: ignore
             except ValueError as e:
                 raise ValueError(
                     'Invalid sample from Stan CSV files, error:\n\t{}\n\t'
                     ' while processing files\n\t{}'.format(
-                        repr(e), '\n\t'.join(mcmc_sample)
+                        repr(e), '\n\t'.join(previous_fit)
                     )
                 ) from e
         else:
             raise ValueError(
-                'MCMC sample must be either CmdStanMCMC object'
-                ' or list of paths to sample Stan CSV files.'
-            )
-        chains = mcmc_fit.chains
-        chain_ids = mcmc_fit.chain_ids
-        if mcmc_fit.metadata.cmdstan_config['save_warmup']:
-            get_logger().warning(
-                'Sample contains saved warmup draws which will be used '
-                'to generate additional quantities of interest.'
+                'Previous fit must be either CmdStanPy fit object'
+                ' or list of paths to Stan CSV files.'
             )
+        if isinstance(fit_object, CmdStanMCMC):
+            chains = fit_object.chains
+            chain_ids = fit_object.chain_ids
+            if fit_object._save_warmup:
+                get_logger().warning(
+                    'Sample contains saved warmup draws which will be used '
+                    'to generate additional quantities of interest.'
+                )
+        elif isinstance(fit_object, CmdStanMLE):
+            chains = 1
+            chain_ids = [1]
+            if fit_object._save_iterations:
+                get_logger().warning(
+                    'MLE contains saved iterations which will be used '
+                    'to generate additional quantities of interest.'
+                )
+        else:  # isinstance(fit_object, CmdStanVB)
+            chains = 1
+            chain_ids = [1]
+
         generate_quantities_args = GenerateQuantitiesArgs(
-            csv_files=sample_csv_files
+            csv_files=fit_csv_files
         )
         generate_quantities_args.validate(chains)
         with MaybeDictToFilePath(data, None) as (_data, _inits):
@@ -1343,7 +1386,7 @@ class CmdStanModel:
                     + ' output is unclear!'
                 )
                 raise RuntimeError(msg)
-            quantities = CmdStanGQ(runset=runset, mcmc_sample=mcmc_fit)
+            quantities = CmdStanGQ(runset=runset, previous_fit=fit_object)
         return quantities
 
     def variational(
@@ -1548,6 +1591,73 @@ class CmdStanModel:
         vb = CmdStanVB(runset)
         return vb
 
+    def log_prob(
+        self,
+        params: Union[Dict[str, Any], str, os.PathLike],
+        data: Union[Mapping[str, Any], str, os.PathLike, None] = None,
+    ) -> pd.DataFrame:
+        """
+        Calculate the log probability and gradient at the given parameter
+        values.
+
+        .. note:: This function is **NOT** an efficient way to evaluate the log
+            density of the model. It should be used for diagnostics ONLY.
+            Please, do not use this for other purposes such as testing new
+            sampling algorithms!
+
+        :param params: Values for all parameters in the model, specified
+            either as a dictionary with entries matching the parameter
+            variables, or as the path of a data file in JSON or Rdump format.
+
+            These should be given on the constrained (natural) scale.
+
+        :param data: Values for all data variables in the model, specified
+            either as a dictionary with entries matching the data variables,
+            or as the path of a data file in JSON or Rdump format.
+
+        :return: A pandas.DataFrame containing columns "lp_" and additional
+            columns for the gradient values. These gradients will be for the
+            unconstrained parameters of the model.
+        """
+
+        if cmdstan_version_before(2, 31, self.exe_info()):
+            raise ValueError(
+                "Method 'log_prob' not available for CmdStan versions "
+                "before 2.31"
+            )
+        with MaybeDictToFilePath(data, params) as (_data, _params):
+            cmd = [
+                str(self.exe_file),
+                "log_prob",
+                f"constrained_params={_params}",
+            ]
+            if _data is not None:
+                cmd += ["data", f"file={_data}"]
+
+            output_dir = tempfile.mkdtemp(prefix=self.name, dir=_TMPDIR)
+
+            output = os.path.join(output_dir, "output.csv")
+            cmd += ["output", f"file={output}"]
+
+            get_logger().debug("Cmd: %s", str(cmd))
+
+            proc = subprocess.run(
+                cmd, capture_output=True, check=False, text=True
+            )
+            if proc.returncode:
+                get_logger().error(
+                    "'log_prob' command failed!\nstdout:%s\nstderr:%s",
+                    proc.stdout,
+                    proc.stderr,
+                )
+                raise RuntimeError(
+                    "Method 'log_prob' failed with return code "
+                    + str(proc.returncode)
+                )
+
+            result = pd.read_csv(output, comment="#")
+            return result
+
     def _run_cmdstan(
         self,
         runset: RunSet,
