@@ -7,12 +7,16 @@ from collections import Counter
 from typing import (
     Any,
     Dict,
+    Generic,
     Hashable,
     List,
     MutableMapping,
+    NoReturn,
     Optional,
     Tuple,
+    TypeVar,
     Union,
+    overload,
 )
 
 import numpy as np
@@ -25,6 +29,7 @@ try:
 except ImportError:
     XARRAY_INSTALLED = False
 
+
 from cmdstanpy.cmdstan_args import Method
 from cmdstanpy.utils import (
     BaseType,
@@ -36,10 +41,14 @@ from cmdstanpy.utils import (
 
 from .mcmc import CmdStanMCMC
 from .metadata import InferenceMetadata
+from .mle import CmdStanMLE
 from .runset import RunSet
+from .vb import CmdStanVB
+
+Fit = TypeVar('Fit', CmdStanMCMC, CmdStanMLE, CmdStanVB)
 
 
-class CmdStanGQ:
+class CmdStanGQ(Generic[Fit]):
     """
     Container for outputs from CmdStan generate_quantities run.
     Created by :meth:`CmdStanModel.generate_quantities`.
@@ -48,7 +57,7 @@ class CmdStanGQ:
     def __init__(
         self,
         runset: RunSet,
-        mcmc_sample: CmdStanMCMC,
+        previous_fit: Fit,
     ) -> None:
         """Initialize object."""
         if not runset.method == Method.GENERATE_QUANTITIES:
@@ -57,7 +66,9 @@ class CmdStanGQ:
                 'found method {}'.format(runset.method)
             )
         self.runset = runset
-        self.mcmc_sample = mcmc_sample
+
+        self.previous_fit: Fit = previous_fit
+
         self._draws: np.ndarray = np.array(())
         config = self._validate_csv_files()
         self._metadata = InferenceMetadata(config)
@@ -85,6 +96,14 @@ class CmdStanGQ:
             # pylint: disable=raise-missing-from
             raise AttributeError(*e.args)
 
+    def __getstate__(self) -> dict:
+        # This function returns the mapping of objects to serialize with pickle.
+        # See https://docs.python.org/3/library/pickle.html#object.__getstate__
+        # for details. We call _assemble_generated_quantities to ensure
+        # the data are loaded prior to serialization.
+        self._assemble_generated_quantities()
+        return self.__dict__
+
     def _validate_csv_files(self) -> Dict[str, Any]:
         """
         Checks that Stan CSV output files for all chains are consistent
@@ -179,7 +198,7 @@ class CmdStanGQ:
         :param concat_chains: When ``True`` return a 2D array flattening all
             all draws from all chains.  Default value is ``False``.
 
-        :param inc_sample: When ``True`` include all columns in the mcmc_sample
+        :param inc_sample: When ``True`` include all columns in the previous_fit
             draws array as well, excepting columns for variables already present
             in the generated quantities drawset. Default value is ``False``.
 
@@ -189,18 +208,32 @@ class CmdStanGQ:
         CmdStanGQ.draws_xr
         CmdStanMCMC.draws
         """
-        if self._draws.shape == (0,):
-            self._assemble_generated_quantities()
-        if (
-            inc_warmup
-            and not self.mcmc_sample.metadata.cmdstan_config['save_warmup']
-        ):
-            get_logger().warning(
-                "Sample doesn't contain draws from warmup iterations,"
-                ' rerun sampler with "save_warmup=True".'
-            )
+        self._assemble_generated_quantities()
+        if inc_warmup:
+            if (
+                isinstance(self.previous_fit, CmdStanMCMC)
+                and not self.previous_fit._save_warmup
+            ):
+                get_logger().warning(
+                    "Sample doesn't contain draws from warmup iterations,"
+                    ' rerun sampler with "save_warmup=True".'
+                )
+            elif (
+                isinstance(self.previous_fit, CmdStanMLE)
+                and not self.previous_fit._save_iterations
+            ):
+                get_logger().warning(
+                    "MLE doesn't contain draws from pre-convergence iterations,"
+                    ' rerun optimization with "save_iterations=True".'
+                )
+            elif isinstance(self.previous_fit, CmdStanVB):
+                get_logger().warning(
+                    "Variational fit doesn't make sense with argument "
+                    '"inc_warmup=True"'
+                )
+
         if inc_sample:
-            cols_1 = self.mcmc_sample.column_names
+            cols_1 = self.previous_fit.column_names
             cols_2 = self.column_names
             dups = [
                 item
@@ -209,20 +242,15 @@ class CmdStanGQ:
             ]
             drop_cols: List[int] = []
             for dup in dups:
-                drop_cols.extend(self.mcmc_sample.metadata.stan_vars_cols[dup])
-
-        start_idx = 0
-        if (
-            not inc_warmup
-            and self.mcmc_sample.metadata.cmdstan_config['save_warmup']
-        ):
-            start_idx = self.mcmc_sample.num_draws_warmup
+                drop_cols.extend(self.previous_fit.metadata.stan_vars_cols[dup])
 
+        start_idx, _ = self._draws_start(inc_warmup)
+        previous_draws = self._previous_draws(True)
         if concat_chains and inc_sample:
             return flatten_chains(
                 np.dstack(
                     (
-                        np.delete(self.mcmc_sample.draws(), drop_cols, axis=1),
+                        np.delete(previous_draws, drop_cols, axis=1),
                         self._draws,
                     )
                 )[start_idx:, :, :]
@@ -232,7 +260,7 @@ class CmdStanGQ:
         if inc_sample:
             return np.dstack(
                 (
-                    np.delete(self.mcmc_sample.draws(), drop_cols, axis=1),
+                    np.delete(previous_draws, drop_cols, axis=1),
                     self._draws,
                 )
             )[start_idx:, :, :]
@@ -268,17 +296,33 @@ class CmdStanGQ:
                 vars_list = [vars]
             else:
                 vars_list = vars
+
             vars_list = list(dict.fromkeys(vars_list))
-        if (
-            inc_warmup
-            and not self.mcmc_sample.metadata.cmdstan_config['save_warmup']
-        ):
-            get_logger().warning(
-                'Draws from warmup iterations not available,'
-                ' must run sampler with "save_warmup=True".'
-            )
-        if self._draws.shape == (0,):
-            self._assemble_generated_quantities()
+
+        if inc_warmup:
+            if (
+                isinstance(self.previous_fit, CmdStanMCMC)
+                and not self.previous_fit._save_warmup
+            ):
+                get_logger().warning(
+                    "Sample doesn't contain draws from warmup iterations,"
+                    ' rerun sampler with "save_warmup=True".'
+                )
+            elif (
+                isinstance(self.previous_fit, CmdStanMLE)
+                and not self.previous_fit._save_iterations
+            ):
+                get_logger().warning(
+                    "MLE doesn't contain draws from pre-convergence iterations,"
+                    ' rerun optimization with "save_iterations=True".'
+                )
+            elif isinstance(self.previous_fit, CmdStanVB):
+                get_logger().warning(
+                    "Variational fit doesn't make sense with argument "
+                    '"inc_warmup=True"'
+                )
+
+        self._assemble_generated_quantities()
 
         gq_cols = []
         mcmc_vars = []
@@ -289,7 +333,7 @@ class CmdStanGQ:
                         gq_cols.append(self.column_names[idx])
                 elif (
                     inc_sample
-                    and var in self.mcmc_sample.metadata.stan_vars_cols
+                    and var in self.previous_fit.metadata.stan_vars_cols
                 ):
                     mcmc_vars.append(var)
                 else:
@@ -298,13 +342,13 @@ class CmdStanGQ:
             gq_cols = list(self.column_names)
             vars_list = gq_cols
 
+        previous_draws_pd = self._previous_draws_pd(mcmc_vars, inc_warmup)
+
         if inc_sample and mcmc_vars:
             if gq_cols:
                 return pd.concat(
                     [
-                        self.mcmc_sample.draws_pd(
-                            vars=mcmc_vars, inc_warmup=inc_warmup
-                        ).reset_index(drop=True),
+                        previous_draws_pd,
                         pd.DataFrame(
                             data=flatten_chains(
                                 self.draws(inc_warmup=inc_warmup)
@@ -315,11 +359,9 @@ class CmdStanGQ:
                     axis='columns',
                 )[vars_list]
             else:
-                return self.mcmc_sample.draws_pd(
-                    vars=mcmc_vars, inc_warmup=inc_warmup
-                )
+                return previous_draws_pd
         elif inc_sample and vars is None:
-            cols_1 = self.mcmc_sample.column_names
+            cols_1 = self.previous_fit.column_names
             cols_2 = self.column_names
             dups = [
                 item
@@ -328,16 +370,13 @@ class CmdStanGQ:
             ]
             return pd.concat(
                 [
-                    self.mcmc_sample.draws_pd(inc_warmup=inc_warmup)
-                    .drop(columns=dups)
-                    .reset_index(drop=True),
+                    previous_draws_pd.drop(columns=dups).reset_index(drop=True),
                     pd.DataFrame(
                         data=flatten_chains(self.draws(inc_warmup=inc_warmup)),
                         columns=self.column_names,
                     ),
                 ],
-                axis='columns',
-                ignore_index=True,
+                axis=1,
             )
         elif gq_cols:
             return pd.DataFrame(
@@ -350,6 +389,24 @@ class CmdStanGQ:
             columns=self.column_names,
         )
 
+    @overload
+    def draws_xr(
+        self: Union["CmdStanGQ[CmdStanMLE]", "CmdStanGQ[CmdStanVB]"],
+        vars: Union[str, List[str], None] = None,
+        inc_warmup: bool = False,
+        inc_sample: bool = False,
+    ) -> NoReturn:
+        ...
+
+    @overload
+    def draws_xr(
+        self: "CmdStanGQ[CmdStanMCMC]",
+        vars: Union[str, List[str], None] = None,
+        inc_warmup: bool = False,
+        inc_sample: bool = False,
+    ) -> "xr.Dataset":
+        ...
+
     def draws_xr(
         self,
         vars: Union[str, List[str], None] = None,
@@ -359,6 +416,9 @@ class CmdStanGQ:
         """
         Returns the generated quantities draws as a xarray Dataset.
 
+        This method can only be called when the underlying fit was made
+        through sampling, it cannot be used on MLE or VB outputs.
+
         :param vars: optional list of variable names.
 
         :param inc_warmup: When ``True`` and the warmup draws are present in
@@ -375,6 +435,11 @@ class CmdStanGQ:
             raise RuntimeError(
                 'Package "xarray" is not installed, cannot produce draws array.'
             )
+        if not isinstance(self.previous_fit, CmdStanMCMC):
+            raise RuntimeError(
+                'Method "draws_xr" is only available when '
+                'original fit is done via Sampling.'
+            )
         mcmc_vars_list = []
         dup_vars = []
         if vars is not None:
@@ -385,7 +450,7 @@ class CmdStanGQ:
             for var in vars_list:
                 if var not in self.metadata.stan_vars_cols:
                     if inc_sample and (
-                        var in self.mcmc_sample.metadata.stan_vars_cols
+                        var in self.previous_fit.metadata.stan_vars_cols
                     ):
                         mcmc_vars_list.append(var)
                         dup_vars.append(var)
@@ -394,17 +459,16 @@ class CmdStanGQ:
         else:
             vars_list = list(self.metadata.stan_vars_cols.keys())
             if inc_sample:
-                for var in self.mcmc_sample.metadata.stan_vars_cols.keys():
+                for var in self.previous_fit.metadata.stan_vars_cols.keys():
                     if var not in vars_list and var not in mcmc_vars_list:
                         mcmc_vars_list.append(var)
         for var in dup_vars:
             vars_list.remove(var)
 
-        if self._draws.shape == (0,):
-            self._assemble_generated_quantities()
+        self._assemble_generated_quantities()
 
-        num_draws = self.mcmc_sample.num_draws_sampling
-        sample_config = self.mcmc_sample.metadata.cmdstan_config
+        num_draws = self.previous_fit.num_draws_sampling
+        sample_config = self.previous_fit.metadata.cmdstan_config
         attrs: MutableMapping[Hashable, Any] = {
             "stan_version": f"{sample_config['stan_version_major']}."
             f"{sample_config['stan_version_minor']}."
@@ -413,8 +477,8 @@ class CmdStanGQ:
             "num_draws_sampling": num_draws,
         }
         if inc_warmup and sample_config['save_warmup']:
-            num_draws += self.mcmc_sample.num_draws_warmup
-            attrs["num_draws_warmup"] = self.mcmc_sample.num_draws_warmup
+            num_draws += self.previous_fit.num_draws_warmup
+            attrs["num_draws_warmup"] = self.previous_fit.num_draws_warmup
 
         data: MutableMapping[Hashable, Any] = {}
         coordinates: MutableMapping[Hashable, Any] = {
@@ -437,11 +501,11 @@ class CmdStanGQ:
                 build_xarray_data(
                     data,
                     var,
-                    self.mcmc_sample.metadata.stan_vars_dims[var],
-                    self.mcmc_sample.metadata.stan_vars_cols[var],
+                    self.previous_fit.metadata.stan_vars_dims[var],
+                    self.previous_fit.metadata.stan_vars_cols[var],
                     0,
-                    self.mcmc_sample.draws(inc_warmup=inc_warmup),
-                    self.mcmc_sample._metadata.stan_vars_types[var],
+                    self.previous_fit.draws(inc_warmup=inc_warmup),
+                    self.previous_fit._metadata.stan_vars_types[var],
                 )
 
         return xr.Dataset(data, coords=coordinates, attrs=attrs).transpose(
@@ -494,7 +558,7 @@ class CmdStanGQ:
         CmdStanMLE.stan_variable
         CmdStanVB.stan_variable
         """
-        model_var_names = self.mcmc_sample.metadata.stan_vars_cols.keys()
+        model_var_names = self.previous_fit.metadata.stan_vars_cols.keys()
         gq_var_names = self.metadata.stan_vars_cols.keys()
         if not (var in model_var_names or var in gq_var_names):
             raise ValueError(
@@ -503,36 +567,40 @@ class CmdStanGQ:
                 + ", ".join(model_var_names | gq_var_names)
             )
         if var not in gq_var_names:
-            return self.mcmc_sample.stan_variable(var, inc_warmup=inc_warmup)
-        else:  # is gq variable
-            if self._draws.shape == (0,):
-                self._assemble_generated_quantities()
-            draw1 = 0
-            if (
-                not inc_warmup
-                and self.mcmc_sample.metadata.cmdstan_config['save_warmup']
-            ):
-                draw1 = self.mcmc_sample.num_draws_warmup
-            num_draws = self.mcmc_sample.num_draws_sampling
-            if (
-                inc_warmup
-                and self.mcmc_sample.metadata.cmdstan_config['save_warmup']
-            ):
-                num_draws += self.mcmc_sample.num_draws_warmup
-            dims = [num_draws * self.chains]
-            col_idxs = self._metadata.stan_vars_cols[var]
-            if len(col_idxs) > 0:
-                dims.extend(self._metadata.stan_vars_dims[var])
+            if isinstance(self.previous_fit, CmdStanMCMC):
+                return self.previous_fit.stan_variable(
+                    var, inc_warmup=inc_warmup
+                )
+            elif isinstance(self.previous_fit, CmdStanMLE):
+                return np.atleast_1d(  # type: ignore
+                    np.asarray(
+                        self.previous_fit.stan_variable(
+                            var, inc_iterations=inc_warmup
+                        )
+                    )
+                )
+            else:
+                return np.atleast_1d(  # type: ignore
+                    np.asarray(self.previous_fit.stan_variable(var))
+                )
 
-            draws = self._draws[draw1:, :, col_idxs]
+        # is gq variable
+        self._assemble_generated_quantities()
 
-            if self._metadata.stan_vars_types[var] == BaseType.COMPLEX:
-                draws = draws[..., ::2] + 1j * draws[..., 1::2]
-                dims = dims[:-1]
+        draw1, num_draws = self._draws_start(inc_warmup)
+        dims = [num_draws * self.chains]
+        col_idxs = self._metadata.stan_vars_cols[var]
+        if len(col_idxs) > 0:
+            dims.extend(self._metadata.stan_vars_dims[var])
+        draws = self._draws[draw1:, :, col_idxs]
 
-            draws = draws.reshape(dims, order='F')
+        if self._metadata.stan_vars_types[var] == BaseType.COMPLEX:
+            draws = draws[..., ::2] + 1j * draws[..., 1::2]
+            dims = dims[:-1]
 
-            return draws
+        draws = draws.reshape(dims, order='F')
+
+        return draws
 
     def stan_variables(self, inc_warmup: bool = False) -> Dict[str, np.ndarray]:
         """
@@ -551,7 +619,7 @@ class CmdStanGQ:
         CmdStanVB.stan_variables
         """
         result = {}
-        sample_var_names = self.mcmc_sample.metadata.stan_vars_cols.keys()
+        sample_var_names = self.previous_fit.metadata.stan_vars_cols.keys()
         gq_var_names = self.metadata.stan_vars_cols.keys()
         for name in gq_var_names:
             result[name] = self.stan_variable(name, inc_warmup)
@@ -561,9 +629,11 @@ class CmdStanGQ:
         return result
 
     def _assemble_generated_quantities(self) -> None:
+        if self._draws.shape != (0,):
+            return
         # use numpy loadtxt
-        warmup = self.mcmc_sample.metadata.cmdstan_config['save_warmup']
-        num_draws = self.mcmc_sample.draws(inc_warmup=warmup).shape[0]
+        _, num_draws = self._draws_start(inc_warmup=True)
+
         gq_sample: np.ndarray = np.empty(
             (num_draws, self.chains, len(self.column_names)),
             dtype=float,
@@ -577,6 +647,77 @@ class CmdStanGQ:
                 )
         self._draws = gq_sample
 
+    def _draws_start(self, inc_warmup: bool) -> Tuple[int, int]:
+        draw1 = 0
+        p_fit = self.previous_fit
+        if isinstance(p_fit, CmdStanMCMC):
+            num_draws = p_fit.num_draws_sampling
+            if p_fit._save_warmup:
+                if inc_warmup:
+                    num_draws += p_fit.num_draws_warmup
+                else:
+                    draw1 = p_fit.num_draws_warmup
+
+        elif isinstance(p_fit, CmdStanMLE):
+            num_draws = 1
+            if p_fit._save_iterations:
+
+                opt_iters = len(p_fit.optimized_iterations_np)  # type: ignore
+                if inc_warmup:
+                    num_draws = opt_iters
+                else:
+                    draw1 = opt_iters - 1
+        else:  # CmdStanVB:
+            draw1 = 1  # skip mean
+            num_draws = p_fit.variational_sample.shape[0]
+            if inc_warmup:
+                num_draws += 1
+
+        return draw1, num_draws
+
+    def _previous_draws(self, inc_warmup: bool) -> np.ndarray:
+        """
+        Extract the draws from self.previous_fit.
+        Return is always 3-d
+        """
+        p_fit = self.previous_fit
+        if isinstance(p_fit, CmdStanMCMC):
+            return p_fit.draws(inc_warmup=inc_warmup)
+        elif isinstance(p_fit, CmdStanMLE):
+            if inc_warmup and p_fit._save_iterations:
+                return p_fit.optimized_iterations_np[:, None]  # type: ignore
+
+            return np.atleast_2d(p_fit.optimized_params_np,)[  # type: ignore
+                :, None
+            ]
+        else:  # CmdStanVB:
+            if inc_warmup:
+                return np.vstack(
+                    [p_fit.variational_params_np, p_fit.variational_sample]
+                )[:, None]
+            return p_fit.variational_sample[:, None]
+
+    def _previous_draws_pd(
+        self, vars: List[str], inc_warmup: bool
+    ) -> pd.DataFrame:
+        if vars:
+            sel: Union[List[str], slice] = vars
+        else:
+            sel = slice(None, None)
+
+        p_fit = self.previous_fit
+        if isinstance(p_fit, CmdStanMCMC):
+            return p_fit.draws_pd(vars or None, inc_warmup=inc_warmup)
+
+        elif isinstance(p_fit, CmdStanMLE):
+
+            if inc_warmup and p_fit._save_iterations:
+                return p_fit.optimized_iterations_pd[sel]  # type: ignore
+            else:
+                return p_fit.optimized_params_pd[sel]
+        else:  # CmdStanVB:
+            return p_fit.variational_sample_pd[sel]
+
     def save_csvfiles(self, dir: Optional[str] = None) -> None:
         """
         Move output CSV files to specified directory.  If files were
@@ -592,3 +733,10 @@ class CmdStanGQ:
         cmdstanpy.from_csv
         """
         self.runset.save_csvfiles(dir)
+
+    @property
+    def mcmc_sample(self) -> Union[CmdStanMCMC, CmdStanMLE, CmdStanVB]:
+        get_logger().warning(
+            "Property `mcmc_sample` is deprecated, use `previous_fit` instead"
+        )
+        return self.previous_fit
