@@ -33,7 +33,7 @@ class CmdStanMLE:
         assert isinstance(
             optimize_args, OptimizeArgs
         )  # make the typechecker happy
-        self._save_iterations = optimize_args.save_iterations
+        self._save_iterations: bool = optimize_args.save_iterations
         self._set_mle_attrs(runset.csv_files[0])
 
     def __repr__(self) -> str:
@@ -152,7 +152,7 @@ class CmdStanMLE:
         return pd.DataFrame(self._all_iters, columns=self.column_names)
 
     @property
-    def optimized_params_dict(self) -> Dict[str, float]:
+    def optimized_params_dict(self) -> Dict[str, np.float64]:
         """
         Returns all estimates from the optimizer, including `lp__` as a
         Python Dict.  Only returns estimate from final iteration.
